require File.expand_path('../../spec_helper', __dir__)
require 'skadategems/dev/skadate'
require 'fileutils'
require 'digest/md5'

module SkadateGems::Dev
  describe Remote::File do
    let(:skadate) { Skadate.new(skadate_local_dir) }

    let(:big_file_path) do
      ENV['SKADATE_BIG_FILE_PATH'] || File.join(skadate_source_dir, 'install/skadate9.sql')
    end

    let(:big_file_name) { File.basename(big_file_path) }

    let(:remote_file) do
      Remote::File.new(skadate.remote, big_file_name, File.size(big_file_path))
    end

    before(:each) do
      remote_install

      symlink = File.join(skadate_remote_dir, big_file_name)
      FileUtils.ln_s(big_file_path, symlink, verbose: verbose?)

      FileUtils.cd(skadate_local_dir, verbose: verbose?)
    end

    describe '#request' do

      it 'yields `http response` that allows us to #read_body as a stream' do
        segments_count = 0

        remote_file.request do |response|
          response.read_body { |_| segments_count += 1 }
        end

        puts "[segments count = #{segments_count}]" if verbose?
        expect(segments_count).to be > 100
      end

      context 'when :md5_checksum is given' do

        context "and remote file's checksum is the same" do
          let(:checksum) { Digest::MD5.file(big_file_path).hexdigest }

          it 'responds with [304 "Not Modified"] and empty body' do
            remote_file.request(md5_checksum: checksum) do |response|
              expect(response).to be_a Net::HTTPNotModified
              response.read_body { |_| raise 'should not have been called!' }
            end
          end
        end

        context "but remote file's checksum is different" do
          let(:checksum) { 'a' * 32 }

          it 'responds with [200 "OK"] and file content as a stream' do
            segments_count = 0

            remote_file.request(md5_checksum: checksum) do |response|
              expect(response).to be_a Net::HTTPOK
              response.read_body { |_| segments_count += 1 }
            end

            puts "[segments count = #{segments_count}]" if verbose?
            expect(segments_count).to be > 100
          end
        end

      end

    end

    describe '#save_as(filename)' do
      let(:local_file_path) { File.join(skadate_local_dir, big_file_name) }

      context 'when there is no local file with a given filename' do
        it 'downloads and saves a remote file' do
          return_value = nil

          expect {
            return_value = remote_file.save_as(local_file_path)
          }.to change { File.exists?(local_file_path) }.to true

          expect(return_value).to be_true

          expect(FileUtils.identical?(big_file_path, local_file_path)).to be_true
        end
      end

      context 'when a given file is already exists' do

        context 'and files are the same' do
          before(:each) do
            FileUtils.cp(big_file_path, local_file_path, verbose: verbose?)
            expect(FileUtils.identical?(big_file_path, local_file_path)).to be_true
          end

          it "doesn't download a remote file" do
            FileUtils.chmod('u-w', local_file_path, verbose: verbose?)
            expect(File.writable?(local_file_path)).to be_false

            return_value = nil

            expect {
              return_value = remote_file.save_as(local_file_path)
            }.to_not change { File.writable?(local_file_path) }

            expect(return_value).to eq false
          end
        end

        context 'but the files are different' do
          before(:each) do
            File.write(local_file_path, 'Oh Noo!')
          end

          it 'downloads and rewrites a local file' do
            return_value = nil

            expect {
              return_value = remote_file.save_as(local_file_path)
            }.to change {
              FileUtils.identical?(local_file_path, big_file_path)
            }.to true

            expect(return_value).to be_true
          end
        end

      end
    end

  end
end
