require 'rubygems'
require 'rubygems/package'
require 'zlib'
require 'fileutils'

module SkadateGems
  module Dev

    class Bundle

      attr_accessor :layout_theme

      # Bundler instance initializer.
      #
      # @param skadate [Skadate]
      def initialize(skadate)
        @source     = skadate.source
        @internal_c = @source.internal_c
        @external_c = @source.external_c
        @userfiles  = @source.userfiles
        @themes_dir = @source.filename('layout/themes/')
      end

      # Filter files by their filename.
      #
      # @param filename [String] absolute path to a file
      # @return [boolean] true if given filename should be bundled
      def should_include?(filename)
        return false if filename.start_with?(@internal_c, @external_c, @userfiles) ||
            filename.end_with?('.tar', '.gz', '.zip', '.swp', '~', 'Thumbs.db', 'desktop.ini')

        if filename.start_with?(@themes_dir) && layout_theme &&
          !filename.start_with?(File.join(@themes_dir, layout_theme))
          return false
        end

        true
      end

      # Create a tar file in memory.
      #
      # @return [StringIO] whose underlying String is the contents of the tar file
      def tar
        return @tar if @tar

        string_io = StringIO.new('')
        Gem::Package::TarWriter.new(string_io) do |tar|

          htaccess = File.join(@source.root, '.htaccess')
          if File.exists?(htaccess)
            tar.add_file '.htaccess', 0644 do |tar_file|
              File.open(htaccess, 'rb') { |file| tar_file.write(file.read) }
            end
          end

          Dir[File.join(@source.root, '**/*')].each do |filename|
            next unless should_include?(filename)

            if File.directory?(filename)
              tar.mkdir (relative_path filename), 0755

              htaccess = File.join(filename, '.htaccess')
              if File.exists?(htaccess)
                filename = htaccess # iteration continues...
              else
                next
              end
            end

            tar.add_file (relative_path filename), 0644 do |tar_file|
              File.open(filename, 'rb') { |file| tar_file.write(file.read) }
            end
          end

          [ (int_c = relative_path @internal_c),
            "#{int_c}/cache",
            "#{int_c}/components",
            "#{int_c}/forms",
            "#{int_c}/lang",
            relative_path(@external_c),
            relative_path(@userfiles),
          ].each { |dir| tar.mkdir dir, 0777 }
        end

        string_io.rewind
        @tar = string_io
      end

      # gzips the underlying string in the given `StringIO`,
      # returning a new `StringIO` representing the compressed file.
      #
      # @param filename [String] optional output filename
      # @return [StringIO|Fixnum]
      def gzip(filename = nil)
        gz = StringIO.new('')
        z = Zlib::GzipWriter.new(gz)
        z.write(tar.string)
        z.close # this is necessary!

        if filename
          File.binwrite(filename, gz.string)
        else
          StringIO.new(gz.string)
        end
      end

    private
      def relative_path(filename)
        filename.sub(/^#{Regexp.escape(@source.root)}\/?/, '')
      end

    end

  end
end
