require 'time'
require 'thor'
require 'json'
require 'skadategems/dev/skadate'
require 'skadategems/dev/remote/clone_logger'
require 'ruby-progressbar'

module SkadateGems::Dev
  module CLI

    class Remote < Thor
      namespace 'remote'

      desc 'ping', 'Send ping request to a remote application'
      long_desc <<-LONGDESC
Send ping request to a remote server's `exec.php` script. \x5
\x5
\x5 Example usage: \x5
\x5
\x5  $> skadate remote ping
\x5
\x5
      LONGDESC

      def ping
        accessor = current.remote.accessor
        script = ExecPHP::ScriptBatch.new { |s| s << 'echo "pong!";' }

        say '=> '
        say accessor.execphp_uri, :cyan

        accessor.exec(script) do |res|
          say '#> '
          say "#{res.code} [#{res.message}]", (res.code == '200' ? :green : :red)

          if res.body.empty?
            say 'Error: empty response body', :red
          elsif res.body != 'pong!'
            say 'Error: wrong response body:', :red
            if res.body.length > 1000
              say res.body[0..984], nil, false
              say ' ... (continued)', :yellow
            else
              say '"' + res.body + '"'
            end
          else
            say '>> '
            say res.body, :white
          end
        end
      end

      desc 'exec FILENAME', 'Execute php scripts remotely'
      long_desc <<-LONGDESC
Send a given php file to a remote application server for execution.

Example usage:
\x5
\x5  $> skadate remote exec path/to/my_script.php
\x5
\x5
      LONGDESC

      def exec(filename)
        puts current.remote.exec { |batch| batch.include_file(filename) }.body
      end

      desc 'config ID [VALUE]', 'Manage remote application configuration'

      option :type, :type => :string,
                    :desc => 'Explicitly specify config value type'

      def config(key_or_id, value = nil)
        key_or_id = key_or_id.to_i if numeric? key_or_id
        config = current.remote.configs[key_or_id]

        say '>> ', (:yellow unless value.nil?), false
        say "##{config.id} ", :white, false
        say '| '
        say config.name, :cyan, false
        say ': '
        say config.value.to_json, (color_for config.value), value.nil?

        unless value.nil?
          value =
            case options[:type]
              when nil
                if value == 'true'
                  true
                elsif value == 'false'
                  false
                elsif numeric? value
                  value.to_i
                else
                  value
                end
              when 'str'
              when 'string'
                value
              when 'int'
              when 'integer'
              when 'number'
                value.to_i
              when 'bool'
              when 'boolean'
                value != 'false' && value != '0'
              when 'json'
                JSON.parse(value)
              else
                raise "unrecognized config value type '#{options[:type]}'"
            end

          say ' => ', :white, false
          say value.to_json, (color_for value)

          say '>> '
          if yes?('perform this update? [yn]')
            config.value = value

            case config.update_status
              when :updated
                say '[updated]', :green
              when :not_modified
                say '[not modified]', :yellow
              else
                say '[error]', :red
            end
          else
            say '[cancelled]', :red
          end
        end
      end

      desc 'clone', 'Clone production-running application'
      long_desc <<-LONGDESC
Clone production-running application locally for development purposes.

To download remote application source files:
\x5
\x5  $> skadate remote clone --sources \x5
\x5

To get a remote application database dump:
\x5
\x5  $> skadate remote clone --database \x5
\x5

Or if you want to make a full clone of a running application including all the data and user uploaded files:
\x5
\x5  $> skadate remote clone --sources --database --userfiles \x5
\x5

Make a note that by default `remote clone --sources` copies only one `layout/theme` that is currently active on a remote server.
If you want to get another theme specify it using the `--layout-theme` switch as following:
\x5
\x5  $> skadate remote clone --layout-theme stencil \x5
\x5
\x5
      LONGDESC

      option :sources, :type => :boolean,
                       :desc => 'Copy application source files'

      option :schema,  :type => :boolean,
                       :desc => 'Copy database schema'

      option :configs, :type => :boolean,
                       :desc => 'Copy application-related configs data'

      option :database, :type => :boolean,
                        :desc => 'Copy the whole application database'

      option :userfiles, :type => :boolean,
                         :desc => 'Copy `userfiles` directory contents'

      option :layout_theme, :type => :string,
                            :banner => 'NAME',
                            :desc => 'Force layout theme'

      IGNORED_DIR_PATHS = %w[
        /$external_c /$internal_c /$userfiles /$userfiles/tmp
         /external_c  /internal_c  /userfiles  /userfiles/tmp
        /cgi-bin
        /layout/themes
        /m/application/cache/smarty_compile
      ]

      IGNORED_FILE_EXTENSIONS = %w[.tar .bz2 .gz .zip .rar .iso]

      # default values, rewritable by ENV['SKADATE_MAX_*_FILE_SIZE']
      MAX_SOURCE_FILE_SIZE = 1024 * 1024 * 2 # ~2 megabytes
      MAX_USER_FILE_SIZE   = 1024 * 1024 * 10 # ~10 megabytes

      # show progressbar when downloading file size exceeds this value
      SHOW_PROGRESS_FILE_SIZE = 1024 * 512 # ~512 kilobytes

      def clone
        if options[:schema] || options[:configs]
          say <<-STDOUT, :red
Remote clone options `--schema`, `--configs` are currently not implemented, please use `--database` for now.
          STDOUT
          return
        end

        if options.size == 0
          say 'Error: `remote clone` options are required', :red
          return help('clone')
        end

        remote = current.remote
        layout_theme = options[:layout_theme]
        @max_file_size = MAX_SOURCE_FILE_SIZE

        dev_logs_dir = ::File.join(current.root_dir, '.dev', 'log')
        FileUtils.mkdir_p(dev_logs_dir) unless ::File.exists?(dev_logs_dir)
        log_filename = ::File.join(dev_logs_dir, "remote-clone[#{Time.now.to_i}].log")

        logger.start(log_filename) do |log|
          if options[:sources]
            log.puts '>> Downloading application sources...', :yellow
            download_source_dir(remote.root, log)

            if layout_theme.nil?
              log.puts('>> Getting active theme name...', :yellow)
              config = remote.configs[1]
              raise <<-ERRMSG unless config.name == 'theme'
unexpected remote application config entry `#{config.name}`, (expected `theme` [config_id: 1])
              ERRMSG
              layout_theme = config.value
            end
          end

          if layout_theme
            log.puts ">> Downloading [/layout/themes/#{layout_theme}/*]...", :yellow
            download_source_dir remote.directory("/layout/themes/#{layout_theme}"), log
          end

          if options[:database]
            log.puts '>> Creating remote database dump...', :yellow
            remote.create_mysql_dump do |remote_file|
              local_filename = File.join(current.root_dir,
                  "remote-db-clone.#{remote_file.basename =~ /\.tar\.gz$/ ? 'tar.gz' : 'sql'}")

              if log.mute?
                remote_file.save_as(local_filename)
              else
                progressbar = nil
                remote_file.save_as(local_filename) do |chunk_size|
                  if progressbar
                    progressbar.progress += chunk_size
                  else
                    progressbar = create_progressbar(remote_file, log.padding, chunk_size)
                  end
                end
              end

              log.puts '>> Remote cache cleanup...', :yellow
            end ? log.puts('>> Done!', :green) : log.puts('>> Warning: remote cache cleanup failed', :red)
          end

          if options[:userfiles]
            log.puts '>> Detecting `$userfiles` directory...', :yellow
            @max_file_size = MAX_USER_FILE_SIZE
            download_source_dir(remote.userfiles, log)
          end

          log.puts '>> Completed!', :green
        end
      end

    protected
      def current
        @current ||= SkadateGems::Dev::Skadate.new(Dir.pwd)
      end

      def numeric?(string)
        string =~ /\A\d+\z/
      end

      def color_for(value)
        if value.is_a? String
          :green
        elsif value.is_a? Fixnum
          :blue
        elsif value.is_a?(TrueClass) || value.is_a?(FalseClass)
          :magenta
        elsif value.is_a?(Array) || value.is_a?(Hash)
          :yellow
        else
          :red
        end
      end

      def logger
        SkadateGems::Dev::Remote::CloneLogger
      end

      # @param directory [Remote::Directory] remote directory entry
      # @param log [Remote::CloneLogger] logger instance
      def download_source_dir(directory, log)
        list = nil
        log.entry(directory) { list = directory.list }
        return unless list

        log.padding += 1

        begin
          list.each do |entry|
            if entry.directory?
              if IGNORED_DIR_PATHS.include?(entry.path)
                log.entry(entry, :skipped, 'skipped by pathname') { }
              else
                download_source_dir(entry, log)
              end
            else
              if IGNORED_FILE_EXTENSIONS.include?(entry.extname)
                log.entry(entry, :skipped, "skipped by (*#{entry.extname})") { }
              elsif entry.size > @max_file_size
                log.entry(entry, :skipped, 'skipped by filesize') { }
              else
                log.entry(entry) do
                  local_filename = File.join(current.root_dir, entry.path)

                  downloaded =
                    if !log.mute? && entry.size > SHOW_PROGRESS_FILE_SIZE
                      progressbar = nil
                      entry.save_as(local_filename) do |chunk_size|
                        if progressbar
                          progressbar.progress += chunk_size
                        else
                          progressbar = create_progressbar(entry, log.padding, chunk_size)
                        end
                      end
                    else
                      entry.save_as(local_filename)
                    end

                  downloaded ? :downloaded : :unchanged
                end
              end
            end
          end
        ensure
          log.padding -= 1
        end
      end

      def create_progressbar(entry, padding, starting_at = 0)
        format =
"#{'  ' * padding}~> %t (#{set_color(entry.friendly_size, :cyan)}) [#{set_color('%p%%', :green)}] [#{set_color('%B', :green)}] %e "

        ProgressBar.create title:  entry.basename,
                           total:  entry.size,
                           smoothing:     0.5,
                           throttle_rate: 0.3,
                           format:     format,
                           progress_mark: '·',
                           starting_at: starting_at
      end

    end

  end
end
