require 'time'
require 'uri'
require 'thor'
require 'execphp'
require 'skadategems/dev/skadate'
require 'skadategems/dev/bundle'
require 'skadategems/dev/cli/remote'

module SkadateGems::Dev
  module CLI

    class Skadate < Thor
      namespace 'skadate'

      desc 'version [ list | [ about ] | number | build ]',
           'Display Skadate software version information'
      long_desc <<-LONGDESC
To detect the version of the software that is located under the current working directory:
\x5
\x5  $> skadate version
\x5

To list all versions supported by the tool:
\x5
\x5  $> skadate version list
\x5
\x5
      LONGDESC
      def version(section = 'about')
        section =
          case section[0]
            when 'l'
              :list
            when 'a'
              :about
            when 'n'
              :number
            when 'b'
              :build
            else
              say "Error: `skadate version` unrecognized info section `#{section}`", :red
              return help('version')
          end

        if section == :list
          current_version = current.source.version

          SoftwareVersion::LIST.each do |version|
            if version == current_version
              say('* ', version_status(current_version)[0], false)
            else
              say '  ', nil, false
            end

            say "skadate-#{version} [#{version.release_date.strftime('%Y-%m-%d')}]"
          end

          return
        end

        version = current.source.version!
        color, status = version_status(version)

        case section
          when :about
            say 'Skadate Software '
            say("#{version} ", :bold)

            if version.codename
              say('(', :white, false)
              say("#{version.codename}", :cyan, false)
              say ') '
            end

            say('[', :white, false)
            say(status, color, false)
            say ']'

            if version.release_date
              say "Release date: #{version.release_date.strftime('%B %e, %Y')}"
            end

            say 'Copyright (c) 2004-2014 Skalfa LLC'

          when :number
            say(version, color)

          when :build
            say(version.build, color)

          else
            raise "unexpected case :#{section}"
        end
      end

      desc 'init', 'Initialize Skadate development project'

      option :remote_url, type: :string,
                          banner: 'URL',
                          aliases: :r,
                          desc: 'Remote application server url'

      def init
        source = current.source

        dev_dir = source.filename('.dev')
        FileUtils.mkdir(dev_dir) unless File.exists?(dev_dir)

        htaccess_file_path = source.filename('.dev/.htaccess')
        unless File.exists?(htaccess_file_path)
          File.write(htaccess_file_path, <<-HTACCESS)
Order Deny,Allow
Deny from all
          HTACCESS
        end

        if options[:remote_url]
          execphp = ExecPHP::ServerSideAccessor.new

          rs_config = ExecPHP::RemoteServer.new(
            URI.join(options[:remote_url], 'exec.php').to_s,
            execphp.access_token
          )

          rs_config.save_as source.filename('.dev/remote-server.json')
          execphp.generate source.filename('exec.php')
        end
      end

      desc 'export [FILENAME]',
           'Export source files for deployment'

      option :layout_theme, type: :string,
                            desc: 'Exclude all themes except a specified one from being bundled',
                            banner: 'NAME',
                            aliases: '-l'

      long_desc <<-LONGDESC
Export current project source files as an archive for deployment.

To save the largest part of disk space,
it's recommended to specify layout theme name using the `--layout-theme` switch,
that will exclude all unnecessary `layout/themes` from being bundled.

For example:
\x5
\x5  $> skadate export --layout-theme stencil
\x5
\x5
      LONGDESC
      def export(filename = nil)
        version = current.source.version!

        filename ||= "./skadate-#{version}-#{Time.now.to_i}.tar.gz"

        bundle = Bundle.new(current)
        bundle.layout_theme = options[:layout_theme] if options[:layout_theme]
        bundle.gzip(filename)
      end

      register Remote, :remote,
               'remote < ping | exec | clone | config | ... > [OPTIONS]',
               'Remote application control'

    protected
      def current
        @current ||= SkadateGems::Dev::Skadate.new(Dir.pwd)
      end

      def version_status(version)
        return :red, 'unknown' unless version.supported?

        list = SoftwareVersion::LIST

        case version
          when list[-1] then [:green, 'latest']
          when list[-2] then [:yellow, 'previous']
          else [:red, 'outdated']
        end
      end
    end

  end
end
