require 'fileutils'
require 'digest/md5'
require 'tempfile'

module SkadateGems
  module Dev

    class Remote

      # Remote file accessor.
      class File

        # @return [String] relative path to a file
        attr_reader :path

        # @return [Fixnum] remote file size in bytes
        attr_reader :size

        def initialize(remote, path, size = nil)
          @remote, @path, @size = remote, path, size
        end

        def directory?
          false
        end

        # @return [String] remote file basename
        def basename
          ::File.basename(@path)
        end

        # @return [String] remote file extension
        def extname
          ::File.extname(@path)
        end

        # @return [String] human readable file size
        def friendly_size
          units = %W(B KiB MiB GiB TiB)

          size, unit = units.reduce(@size.to_f) do |(fsize, _), type|
            fsize > 512 ? [fsize / 1024, type] : (break [fsize, type])
          end

          "#{size > 9 || size.modulo(1) < 0.1 ? '%d' : '%.1f'} %s" % [size, unit]
        end

        # Request remote file to read it's contents.
        # @yield [response] passes a http response to the block
        def request(md5_checksum: nil, &block)
          path = self.path.dup
          path[0] = '' if path[0] == '/'

          batch = ScriptBatch.new

          batch << <<-PHPSCRIPT
$filename = DIR_SITE_ROOT . '#{path.gsub(/['\\]/, '\\\\\0')}';

if (!file_exists($filename)) {
  header('HTTP/1.1 404 Not Found');
  exit;
}

$md5_checksum = md5_file($filename);
          PHPSCRIPT

          if md5_checksum
            raise ':md5_checksum must be a String' unless md5_checksum.is_a? String

            batch << <<-PHPSCRIPT
if ($md5_checksum === '#{md5_checksum.gsub(/['\\]/, '\\\\\0')}') {
  header('HTTP/1.1 304 Not Modified');
  exit;
}
            PHPSCRIPT
          end

          batch << <<-PHPSCRIPT
header('Content-Description: File Transfer');
header('Content-Type: application/octet-stream');
header('Content-Disposition: attachment; filename=' . basename($filename));
header('Content-Length: ' . filesize($filename));
header('X-Content-MD5-Checksum: ' . $md5_checksum);
ob_clean();
flush();
readfile($filename);
exit;
          PHPSCRIPT

          @remote.accessor.exec(batch) do |response|
            if response.is_a?(Net::HTTPNotModified)
              block.call(response)
            elsif response.is_a?(Net::HTTPOK)
              if response['content-disposition'] == "attachment; filename=#{basename}" &&
                  response['x-content-md5-checksum']
                block.call(response)
              else
                raise 'unexpected response headers'
              end
            else
              raise "[#{response.code} \"#{response.message}\"]"
            end
          end
        end

        # Download and save a remote file as a given filename.
        # @param filename [String] filename to save file as
        # @yield [chunk_size] on each segment receive
        # @yieldparam chunk_size [Fixnum] received segment size
        def save_as(filename)
          current_checksum = if ::File.exists?(filename)
            Digest::MD5.file(filename).hexdigest
          end

          request(md5_checksum: current_checksum) do |response|
            if response.is_a?(Net::HTTPNotModified)
              return false
            elsif response.is_a?(Net::HTTPOK)
              @size = response['content-length'].to_i if @size.nil?
              new_checksum = response['x-content-md5-checksum']
              md5 = Digest::MD5.new

              begin
                tempfile = Tempfile.create([basename, extname])
                response.read_body do |chunk|
                  tempfile << chunk
                  md5 << chunk
                  yield chunk.size if block_given?
                end
                tempfile.close

                if md5.hexdigest != new_checksum
                  raise "md5 checksum mismatch:\n  [#{md5.hexdigest}]\n  [#{new_checksum}]"
                end

                parent_dir = ::File.dirname(filename)
                FileUtils.mkdir_p(parent_dir) unless ::File.exists?(parent_dir)

                FileUtils.mv(tempfile.path, filename)
                FileUtils.chmod(0644, filename)

                return true
              ensure
                tempfile.close unless tempfile.closed?
              end
            else
              raise "unexpected response #{response}"
            end
          end
        end

      end

    end

  end
end
