require 'date'

module SkadateGems
  module Dev

    # Represents Skadate software version
    class SoftwareVersion
      LIST, MAP = [], {}

      attr_reader :major
      attr_reader :minor
      attr_reader :build
      attr_reader :release_date
      attr_reader :codename

      # @param version [String] string version representation
      # @param release_date [Date] version release date
      # @param codename [String] software version codename
      def initialize(version, release_date = nil, codename = nil)
        @major, @minor, @build = version.split('.', 3).map { |n| n.to_i }
        @release_date = release_date
        @codename = codename
      end

      # @return [String] version string representation
      def to_s
        "#{@major}.#{@minor}.#{@build}"
      end

      # Versions equality operator
      def == (other)
        return other == to_s if other.instance_of?(String)

        return false unless other.instance_of?(SoftwareVersion)

        @major == other.major &&
        @minor == other.minor &&
        @build == other.build
      end

      # @return [boolean] true if current version is supported by `skadategems-dev`
      def supported?
        MAP.key?(to_s)
      end

      # Raised in case of software version compatibility error.
      # @see Dev::SourceController#version!
      class CompatibilityError < StandardError; end
    end

    # Skadate software supported versions.
    [ ['9.0.2500', Date.new(2012,  1, 30)],
      ['9.0.2555', Date.new(2012,  5, 23)],
      ['9.2.2657', Date.new(2012, 10, 12)],
      ['9.2.2960', Date.new(2013,  4, 16), 'Master'],
      ['9.3.3144', Date.new(2013, 10,  8), 'Master II']
    ].each do |params|
      version = SoftwareVersion.new(*params)
      SoftwareVersion::LIST << version
      SoftwareVersion::MAP[params[0]] = version
    end

    SoftwareVersion::LIST.freeze
    SoftwareVersion::MAP.freeze
  end
end
