require 'skadategems/dev/software_version'

module SkadateGems
  module Dev

    # Skadate software source controller
    class SourceController

      # @return [String] source root directory
      attr_reader :root

      # @param root_dir [String] source root directory
      def initialize(root_dir)
        @root = root_dir
      end

      # Get full path to a file from the current source model.
      # @param relative_path [String] relative path to a file
      # @return [String] absolute path to a file
      def filename(relative_path)
        File.join(@root, relative_path)
      end

      # @return [String] full path to a `internals/Header.inc.php` file
      def header_inc_php
        filename 'internals/Header.inc.php'
      end

      # Silently tries to get software version object.
      # @return [SoftwareVersion] or nil in case of file reading or parsing error
      def version
        return @version if @version

        begin
          string_def = parse_version_definition
          if SoftwareVersion::MAP.key?(string_def)
            @version = SoftwareVersion::MAP[string_def]
          else
            @version = SoftwareVersion.new(string_def)
          end
        rescue Errno::ENOENT, ParseError
          nil
        end
      end

      # @return [SoftwareVersion] software version object
      # @raise [IOError] if `internals/Header.inc.php` can not be found
      # @raise [SoftwareVersion::IncompatibilityError]
      #   if detected version is not compatible with current version of `skadategems-dev`
      def version!
        if @version.nil?
          string_def = parse_version_definition

          if SoftwareVersion::MAP.key?(string_def)
            @version = SoftwareVersion::MAP[string_def]
          else
            raise SoftwareVersion::CompatibilityError
          end
        end

        @version
      end

      # Search and parse for a version definition within the `#header_inc_php` file.
      # @raise [SourceController::ParseError] if the version definition can not be parsed
      # @private use #version or #version! to gather a software version
      def parse_version_definition
        File.open(header_inc_php) do |file|
          /define\(\s?(['"])PACKAGE_VERSION\1\s?,\s?(['"])(\d{1,2}.\d{1,2}.\d{4})\2\s?\)\s?;/.match(file.read)
        end

        $3 or raise ParseError
      end

      # @return [String] full path to a `internal_c/` directory
      def internal_c
        filename (version!.build < 2960 ? '$internal_c' : 'internal_c')
      end

      # @return [String] full path to a `external_c/` directory
      def external_c
        filename (version!.build < 2960 ? '$external_c' : 'external_c')
      end

      # @return [String] full path to a `userfiles/` directory
      def userfiles
        filename (version!.build < 2960 ? '$userfiles' : 'userfiles')
      end

      # @return [String] full path to a `internals/config.php` file
      def config_php
        filename ("internals/#{(version!.build < 2960 ? '$config.php' : 'config.php')}")
      end

      # Raised in case of source file parse error.
      # @see SourceController#version!
      class ParseError < StandardError; end
    end

  end
end
